//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Group/SelectionProperty.h
//! @brief     Defines class SelectionProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_GROUP_SELECTIONPROPERTY_H
#define BORNAGAIN_GUI_MODEL_GROUP_SELECTIONPROPERTY_H

#include "GUI/Model/Group/SelectionDescriptor.h"

// #baMigration docu
template <typename T> class SelectionProperty {
public:
    ~SelectionProperty() { delete m_p; }

    //! Initializes the selection with the first type in the catalog types.
    //! For initializing a newly created item when creating via Catalog
    template <typename Catalog, typename... ArgsForCreation>
    void init(const QString& label, const QString& tooltip, const QString& persistentTag,
              ArgsForCreation... argsForCreation)
    {
        m_persistentTag = persistentTag;
        m_descriptor.template init<Catalog>(label, tooltip, &m_p, argsForCreation...);
        m_p = Catalog::create(Catalog::types()[0], argsForCreation...);
    }

    //! Initializes the selection with the first type in the catalog types.
    //! For initializing a newly created item in a separate initializer method
    template <typename Catalog>
    void initWithInitializer(const QString& label, const QString& tooltip,
                             const QString& persistentTag,
                             std::function<void(T, const T)> initializer)
    {
        m_initializer = initializer;
        m_persistentTag = persistentTag;
        m_descriptor.template init<Catalog>(label, tooltip, &m_p);
        T t = Catalog::create(Catalog::types()[0]);
        if (t != nullptr)
            initializer(t, m_p);
        m_p = t;
    }

    //! Initializes the selection with the first type of the given types subset.
    //! For initializing a newly created item in a separate initializer method.
    //! Only a subset of types shall be used
    template <typename Catalog>
    void initWithInitializer(const QString& label, const QString& tooltip,
                             const QString& persistentTag,
                             const QVector<typename Catalog::Type>& types,
                             std::function<void(T, const T)> initializer)
    {
        m_initializer = initializer;
        m_persistentTag = persistentTag;
        m_descriptor.template init<Catalog>(label, tooltip, &m_p);
        T t = Catalog::create(types[0]);
        if (t != nullptr)
            initializer(t, m_p);
        m_p = t;
    }

    SelectionDescriptor<T> descriptor() const { return m_descriptor; }

    operator SelectionDescriptor<T>() const { return m_descriptor; }
    T operator->() const { return m_p; }

    //! Directly set the new item.
    void set(T t, bool callInitializer = false)
    {
        if (callInitializer && m_initializer)
            m_initializer(t, m_p);
        delete m_p;
        m_p = t;
    }

    template <typename S, typename... ArgsForConstructor>
    S* set(ArgsForConstructor... argsForConstructor)
    {
        S* s = new S(argsForConstructor...);
        if (s != nullptr && m_initializer)
            m_initializer(s, m_p);
        m_p = s;
        return s;
    }

    T get() const { return m_p; }

    QString persistentTag() const { return m_persistentTag; }

private:
    QString m_persistentTag;
    SelectionDescriptor<T> m_descriptor;

    T m_p = nullptr;
    std::function<void(T, const T)> m_initializer;
    std::function<T*()> m_creator;
};


#endif // BORNAGAIN_GUI_MODEL_GROUP_SELECTIONPROPERTY_H
